/*
 * espeak.c - SubAloud subtitles reader (espeak/mbrola module)
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <espeak/speak_lib.h>
#include <glib.h>

#include "../module.h"

int espeak_samplerate;
int espeak_wavesamples;
short *espeak_wave;
int espeak_wavesize;
int espeak_is_mbrola;
char *espeak_data_path;
int espeak_synth_flags=0;
int espeak_mbrola_mode;

static void append_wave(short *wav,int numsamples)
{
    if (!espeak_wave) {
	espeak_wavesize = 65536;
	while (espeak_wavesize < 2*numsamples) espeak_wavesize *= 2;
	espeak_wave=malloc(2*espeak_wavesize);
	espeak_wavesamples=0;
    }
    else if (espeak_wavesamples + numsamples > espeak_wavesize) {
	while (espeak_wavesamples + 2* numsamples > espeak_wavesize) {
	    espeak_wavesize *= 2;
	}
	espeak_wave=realloc(espeak_wave,2*espeak_wavesize);
    }
    memcpy(espeak_wave+espeak_wavesamples,wav,2*numsamples);
    espeak_wavesamples += numsamples;
}


static int SynthCallback(short *wav, int numsamples, espeak_EVENT *events)
{
    char fname[210];
    if(!wav || espeak_is_mbrola) {
	return 0;
    }

    while(events->type != 0)
    {
	if(events->type == espeakEVENT_SAMPLERATE) {
	    espeak_samplerate = events->id.number;
	}
	events++;
    }
	
    if(numsamples > 0)
    {
	append_wave(wav,numsamples);
    }
    return 0;
}

static int init_espeak(void)
{
    espeak_samplerate = espeak_Initialize(AUDIO_OUTPUT_SYNCHRONOUS,0,espeak_data_path,0);
    espeak_SetSynthCallback(SynthCallback);
}

static char *get_mbrola_phonemes(char *str)
{
    char *phones;
    size_t len;
    FILE *f=open_memstream(&phones,&len);
    espeak_SetPhonemeTrace(16,f);
    espeak_Synth(str,strlen(str)+1,0,POS_CHARACTER,0,espeak_synth_flags,NULL,NULL);
    espeak_Synchronize();
    fclose(f);
    if (phones && !*phones) {
	free(phones);
	return NULL;
    }
    return phones;
}

static void find_mbrola_voice(char *v)
{
	static char *vpaths[]={
		"/usr/share/mbrola/voices/%s",
		"/usr/share/mbrola/%s/%s",
		NULL};
	char buf[256];
	int i;
	if (strlen(v)!=3) return;
	for (i=0;vpaths[i];i++) {
		sprintf(buf,vpaths[i],v,v);
		if (g_file_test(buf,G_FILE_TEST_IS_REGULAR)) {
			strcpy(mbrola_voice,buf);
			break;
		}
	}
}


char *init_speaker_module(char **error)
{
    if (!module_voice) {
	*error=strdup("Module voice not set");
	return NULL;
    }
    if (preproc_type == PREPROC_PREPROC) {
	if (!init_preproc(error)) {
	    return NULL;
	}
    }
    else if (preproc_type != PREPROC_NONE) {
	preproc_type = PREPROC_STANDARD;
    }
    if (!strncmp(module_voice,"mb-",3) || strstr(module_voice,"-mbrola-")) {
	espeak_is_mbrola = 1;
	if (!mbrola_voice[0]) {
	    if (!strncmp(module_voice,"mb-",3)) {
		find_mbrola_voice(module_voice+3);
	    }
	    if (!mbrola_voice[0]) {
		*error=g_strdup("Cannot find mbrola voice");
		return NULL;
	    }
	}
    }
    init_espeak();
    if(espeak_SetVoiceByName(module_voice) != EE_OK) {
	espeak_VOICE voice_select;
	memset(&voice_select,0,sizeof(voice_select));
	voice_select.languages = module_voice;
	if(espeak_SetVoiceByProperties(&voice_select) != EE_OK) {
	    *error=g_strdup_printf("Error loading voice '%s'\n",module_voice);
	    return NULL;
	}
    }
    if (espeak_is_mbrola && preproc_type != PREPROC_PREPROC) {
	espeak_mbrola_mode=1;
    }
    if (espeak_mbrola_mode) {
	return strdup("trans=mbrola\nwave\nrwave\n");
    }
    else {
	return strdup("trans=private\nwave\n");
    }
}

char *get_phones(char *str)
{
    char *c;
    gunichar z;
    int fnd=0;
    if (strstr(str,">>>") || strstr(str,"<<<")) return NULL;
    if (preproc_type == PREPROC_STANDARD) {
	char *s,*rs;
	s=simplify_text(str);
	if (!espeak_is_mbrola) return s;
	if (!s) return NULL;
	rs=get_mbrola_phonemes(s);
	g_free(s);
	return rs;
    }
    for (c=str;*c;) {
	z=g_utf8_get_char(c);
	if (g_unichar_isalnum(z)) {
	    fnd=1;
	    break;
	}
	c=g_utf8_next_char(c);
    }
    if (!fnd) return NULL;
    if (!espeak_mbrola_mode) return g_strdup(str);
    return get_mbrola_phonemes(str);
}



// str - mbrola phonemes

short *get_timed_wave(char *str,int *len,int *offset,int *freq,double rate)
{
    short *wave=NULL;
    if (!str || !*str) return NULL;
    wave=mbrola_get_wave(mbrola_voice,str,mbrola_pitch,1.0/rate,len,offset,freq);
    *offset=0;
    return wave;
}

    

short *get_wave(char *str,int *len,int *offset,int *freq)
{
	short *wave=NULL;
	char *err,*conv=NULL,*pho=NULL,*s;
	if (!str || !*str) return NULL;
	int flags = 0;
	if (espeak_mbrola_mode) {
	    return get_timed_wave(str,len,offset,freq,1.0);
	}
	if (preproc_type == PREPROC_PREPROC) {
	    conv=preproc_string(str,1,1);
	    if (!conv) return NULL;
	}
	if (espeak_is_mbrola) {
	    pho=get_mbrola_phonemes(conv?conv:str);
	    wave=get_timed_wave(pho,len,offset,freq,1.0);
	    if (conv) g_free(conv);
	    free(pho);
	    return wave;
	}
    s=conv?conv:str;
    espeak_SetPhonemeTrace(0,NULL);
    espeak_Synth(s,strlen(s)+1,0,POS_CHARACTER,0,espeak_synth_flags,NULL,NULL);
    if (conv) g_free(conv);
    espeak_Synchronize();
    wave=espeak_wave;
    espeak_wave=NULL;
    *len=espeak_wavesamples;
    *freq=espeak_samplerate;
    *offset=0;
    return wave;
}
	
