/*
 * modpico.c - SubAloud subtitles reader (svox pico module)
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <picoapi.h>

#include "../module.h"

static pico_System picoSystem;
static pico_Resource picoTaResource;
static pico_Resource picoSgResource;
static pico_Engine picoEngine;
static pico_Char *picoInp;

#define PICO_MEM_SIZE			(10000000)
#define MAX_OUTBUF_SIZE 128

char PicoLingwarePath[PATH_MAX]="/usr/share/pico/lang/";

static const int PICO_SAMPLE_RATE = 16000;
static int picoVoiceIndex;

static const char *picoInternalTaLingware[] = {
	"en-US_ta.bin",
	"en-GB_ta.bin",
	"de-DE_ta.bin",
	"es-ES_ta.bin",
	"fr-FR_ta.bin",
	"it-IT_ta.bin"
};

static const char *picoInternalSgLingware[] = {
	"en-US_lh0_sg.bin",
	"en-GB_kh0_sg.bin",
	"de-DE_gl0_sg.bin",
	"es-ES_zl0_sg.bin",
	"fr-FR_nk0_sg.bin",
	"it-IT_cm0_sg.bin"
};

static struct picoVoice {
    char *name;
    char *lang;
    char *ulang;
} pico_voices[] = {
	{"Samantha", "en", "en-US"},
	{"Serena", "en", "en-GB"},
	{"Sabrina", "de", "de-DE"},
	{"Isabel", "es", "es-ES"},
	{"Virginie", "fr", "fr-FR"},
	{"Silvia", "it", "it-IT"}
};

static int getVoiceIndex(void)
{
    int i;
    if (!strcasecmp(module_voice,"us")) return 0;
    if (!strcasecmp(module_voice,"gb")) return 1;
    for (i=0;i<6;i++) {
	if (!strcasecmp(module_voice,pico_voices[i].name) ||
	    !strcasecmp(module_voice,pico_voices[i].ulang) ||
	    !strcasecmp(module_voice,pico_voices[i].lang)) {
		return i;
	}
    }
    return -1;
}

int pico_init_voice(char **error)
{
	int ret;
	pico_Retstring outMessage;
	pico_Char picoTaFileName[PICO_MAX_DATAPATH_NAME_SIZE +
				 PICO_MAX_FILE_NAME_SIZE];
	pico_Char picoSgFileName[PICO_MAX_DATAPATH_NAME_SIZE +
				 PICO_MAX_FILE_NAME_SIZE];
	pico_Char picoTaResourceName[PICO_MAX_RESOURCE_NAME_SIZE];
	pico_Char picoSgResourceName[PICO_MAX_RESOURCE_NAME_SIZE];

	/* Load the text analysis Lingware resource file.   */
	strcpy((char *)picoTaFileName, PicoLingwarePath);
	strcat((char *)picoTaFileName,
	       (const char *)picoInternalTaLingware[picoVoiceIndex]);
	if ((ret =
	     pico_loadResource(picoSystem, picoTaFileName, &picoTaResource))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error=g_strdup_printf("Cannot load TA Lingware resource file (%i): %s",
		    ret,
		    outMessage);
		return 0;
	}

	/* Load the signal generation Lingware resource file.   */
	strcpy((char *)picoSgFileName, PicoLingwarePath);
	strcat((char *)picoSgFileName,
	       (const char *)picoInternalSgLingware[picoVoiceIndex]);
	if ((ret =
	     pico_loadResource(picoSystem, picoSgFileName, &picoSgResource))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error=g_strdup_printf(
		    "Cannot load SG Lingware resource file (%i): %s\n", ret,
		    outMessage);
		return 0;
	}

	/* Get the text analysis resource name.     */
	if ((ret = pico_getResourceName(picoSystem, picoTaResource,
					(char *)picoTaResourceName))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error=g_strdup_printf(
		    "Cannot get TA resource name (%i): %s", ret, outMessage);
		return 0;
	}

	/* Get the signal generation resource name. */
	if ((ret = pico_getResourceName(picoSystem, picoSgResource,
					(char *)picoSgResourceName))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error=g_strdup_printf(
		    "Cannot get SG resource name (%i): %s", ret, outMessage);
		return 0;
	}

	/* Create a voice definition.   */
	if ((ret = pico_createVoiceDefinition(picoSystem,
					      (const pico_Char *)
					      pico_voices[picoVoiceIndex].name))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error=g_strdup_printf(
		    "Cannot create voice definition (%i): %s", ret,
		    outMessage);
		return 0;
	}

	/* Add the text analysis resource to the voice. */
	if ((ret = pico_addResourceToVoiceDefinition(picoSystem,
						     (const pico_Char *)
						     pico_voices[picoVoiceIndex].
						     name,
						     picoTaResourceName))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error=g_strdup_printf(
		    "Cannot add TA resource to the voice (%i): %s",
		    ret, outMessage);
		return 0;
	}

	/* Add the signal generation resource to the voice. */
	if ((ret = pico_addResourceToVoiceDefinition(picoSystem,
		    (const pico_Char *)
		    pico_voices[picoVoiceIndex].name,
		    picoSgResourceName))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error=g_strdup_printf(
		    "Cannot add SG resource to the voice (%i): %s",
		    ret, outMessage);
		return 0;
	}

	return 1;
}

char *init_speaker_module(char **error)
{
    int ret;
    pico_Retstring outMessage;
    
    if (!module_voice) {
	*error=strdup("Module voice not set");
	return NULL;
    }
    if (preproc_type == PREPROC_PREPROC) {
	if (!init_preproc(error)) {
	    return NULL;
	}
    }
    else if (preproc_type != PREPROC_NONE) {
	preproc_type = PREPROC_STANDARD;
    }

    picoVoiceIndex=getVoiceIndex();
    if (picoVoiceIndex < 0) {
	*error=strdup("Voice unknown to module");
	return NULL;
    }
    if (datapath[0]) {
	if (datapath[strlen(datapath)-1] != '/') {
	    strcat(datapath,"/");
	}
	strcpy(PicoLingwarePath,datapath);
    }
    void *pmem = g_malloc(PICO_MEM_SIZE);
    if ((ret = pico_initialize(pmem, PICO_MEM_SIZE, &picoSystem))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error = g_strdup_printf("Cannot initialize (%i): %s\n",
					       ret, outMessage);
		g_free(pmem);
		return NULL;
	}
    if (!pico_init_voice(error)) {
	return NULL;
    }
    if ((ret = pico_newEngine(picoSystem,
				  (const pico_Char *)pico_voices[picoVoiceIndex].name,
				  &picoEngine))) {
		pico_getSystemStatusMessage(picoSystem, ret, outMessage);
		*error = g_strdup_printf(
		    "Cannot create a new pico engine (%i): %s\n",
					       ret, outMessage);
		return NULL;
    }
    return strdup("trans=private\nwave\n");
}

char *get_phones(char *str)
{
    char *c;
    gunichar z;
    int fnd=0;
    if (strstr(str,">>>") || strstr(str,"<<<")) return NULL;
    if (preproc_type == PREPROC_STANDARD) return simplify_text(str);
    for (c=str;*c;) {
	z=g_utf8_get_char(c);
	if (g_unichar_isalnum(z)) {
	    fnd=1;
	    break;
	}
	c=g_utf8_next_char(c);
    }
    if (!fnd) return NULL;
    return g_strdup(str);
}

short *get_timed_wave(char *str,int *len,int *offset,int *freq,double rate)
{
    return NULL;
}

short *get_wave(char *str,int *len,int *offset,int *freq)
{
    pico_Int16 bytes_sent, bytes_recv, text_remaining, out_data_type;
    int ret, getstatus;
    short outbuf[MAX_OUTBUF_SIZE];
    pico_Retstring outMessage;
    
    char *wave=NULL;
    int wave_size,wave_length=0;
    
    if (!str || !*str) return NULL;
    if (preproc_type == PREPROC_PREPROC) {
	str=preproc_string(str,1,1);
	if (!str) return NULL;
    }
    pico_Char *buf = str;

    text_remaining = strlen(buf) + 1;
    
    while (text_remaining>0) {
	if ((ret = pico_putTextUtf8(picoEngine, buf, text_remaining,
					    &bytes_sent))) {
	    pico_getSystemStatusMessage(picoSystem, ret,outMessage);
	    fprintf(stderr,"Cannot put Text (%i): %s\n", ret,outMessage);
	    return NULL;
	}
	text_remaining -= bytes_sent;
	buf += bytes_sent;
	do {
	    getstatus = pico_getData(picoEngine, (void *)outbuf,
						 MAX_OUTBUF_SIZE, &bytes_recv,
						 &out_data_type);
	    if ((getstatus != PICO_STEP_BUSY)
		&& (getstatus != PICO_STEP_IDLE)) {
				pico_getSystemStatusMessage(picoSystem,
							    getstatus,
							    outMessage);
				fprintf(stderr,"Cannot get Data (%i): %s\n",
				    getstatus, outMessage);
				if (preproc_type == PREPROC_PREPROC) g_free(str);
				return NULL;
	    }
	    if (bytes_recv) {
		if (!wave) {
		    wave=malloc(wave_size=8192);
		    wave_length=0;
		}
		else while (wave_length + bytes_recv > wave_size) {
		    wave_size *= 2;
		    wave = realloc(wave, wave_size);
		}
		memcpy(wave + wave_length, outbuf, bytes_recv);
		wave_length += bytes_recv;
	    }
	    
	} while (PICO_STEP_BUSY == getstatus);
    }
    if (preproc_type == PREPROC_PREPROC) g_free(str);
    *freq = PICO_SAMPLE_RATE;
    *len=wave_length/2;
    *offset=0;
    return (short *)wave;
}
