/*
 * simplify.c - simple text normalizer for TTS
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *	       <http://www.gnu.org/licenses/>.
 */

//#include <string.h>
#include <glib.h>


static int is_dash(gunichar ch)
{
    return (ch=='-' || (ch >= 0x2012 && ch <= 0x2015));
}

static int is_apos(gunichar ch)
{
    return (ch == '\'' || ch=='`' || ch == 0x2018 ||
    ch == 0x2019 || ch == 0x201B);
}



static char *start_readable(char *str,int *ldash)
{
    int last_dash=0;
    gunichar g;
    for (;*str;str=g_utf8_next_char(str)) {
    g=g_utf8_get_char(str);
    if (!g) break;
    
    if (g_unichar_isalnum(g)) break;
    if (g == '$' || g == 0xa1 || g == 0xbf) break;
    if (g == '.' && *str && (*str >= '0' && *str <= '9')) {
	last_dash='0';
	break;
    }
    else if (is_dash(g)) last_dash='-';
    else last_dash=0;
    }
    if (!*str) return 0;
    char *c=str;
    for (;*c;) {
    g=g_utf8_get_char(c);
    if (!g) return NULL;
    if (g_unichar_isalnum(g)) {
	*ldash=last_dash;
	return str;
    }
    c=g_utf8_next_char(c);
    }
    return NULL;
}


#define PUNCT_QUESTION 16
#define PUNCT_EXCLAM 8
#define PUNCT_DOT 4
#define PUNCT_COMMA 2
#define PUNCT_BLANK 1
#define PUNCT_MASK 31

#define CH_ALPHA 1
#define CH_ASCIINUM 2
#define CH_NUMERIC 4
#define CH_ASCIIALPHA 8
#define CH_READABLE 16
#define CH_BLANK 32
#define CH_PUNCT 64

#define c_alnum(x) ((x) & (CH_ALPHA | CH_NUMERIC))
#define c_alpha(x) ((x) & CH_ALPHA)
#define c_ascalnum(x) ((x) & (CH_ASCIIALPHA | CH_ASCIINUM))
#define c_ascnum(x) ((x) & CH_ASCIINUM)
#define c_readable(x) ((x) & (CH_READABLE | CH_ALPHA | CH_NUMERIC))
#define c_punct(x) ((x) & CH_PUNCT)

static gunichar commas[]={
    ',',';',':','(',')','[',']','{','}',0x2026,0};

static gunichar punctos[]={
    '.','?','!',0};
    
static gunichar readables[]={
    0x23,0x24,0x25,0x26,0x2a,0x2b,0x2f,
    0x3c,0x3d,0x3e,0x5c,0x40,0x7e,
    0xa2,0xa3,0xa4,0xa5,0xa7,0xa9,0xae,
    0xb0,0xb1,0xb5,0xd7,0xf7,0};
    
    
static int get_class(gunichar g)
{
    int i;
    if (g >= '0' && g <= '9') {
    return CH_NUMERIC | CH_ASCIINUM;
    }
    if ((g >= 'a' && g <= 'z') || (g>='A' && g <= 'Z')) {
    return CH_ALPHA | CH_ASCIIALPHA;
    }
    if (g_unichar_isalnum(g)) {
    if (g_unichar_isdigit(g)) return CH_NUMERIC;
    return CH_ALPHA;
    }
    for (i=0;commas[i];i++) if (g == commas[i]) return CH_PUNCT;
    for (i=0;punctos[i];i++) if (g == punctos[i]) return CH_PUNCT;
    for (i=0;readables[i];i++) if (g == readables[i]) return CH_READABLE;
    return 0;
}
    
    
    

static gunichar get_char(char **str)
{
    gunichar g;
    if (**str=='.' && (*str)[1]=='.' && (*str)[2]=='.') {
	g=0x2026;
	(*str)+=3;
	return g;
    }
    if ((**str == '<' || **str=='>') && (*str)[1] == **str) {
	str+=2;
	return '"';
    }
    g=g_utf8_get_char(*str);
    if (g) *str=g_utf8_next_char(*str);
    if (!g) return 0;
    if (is_dash(g)) return '-';
    if (is_apos(g)) return '\'';
    if (g_unichar_isspace(g)) return ' ';
    return g;
}

static int get_punct(gunichar g)
{
    int i;
    if (g=='?') return PUNCT_QUESTION;
    if (g=='!') return PUNCT_EXCLAM;
    if (g=='.') return PUNCT_DOT;
    for (i=0;commas[i];i++) {
	if (g == commas[i]) return PUNCT_COMMA;
    }
    return 0;
}

char *simplify_text(char *str)
{
    GString *gs;
    
    int last_char,this_char,next_char;
    int last_class,this_class,next_class;
    int last_dash,last_punct;
    
    str=start_readable(str,&last_dash);
    if (!str || !*str) return NULL;
    
    
    
    this_char=get_char(&str);
    if (!this_char) {
	return NULL;
    }
    this_class=get_class(this_char);
    next_char=get_char(&str);
    next_class=get_class(next_char);
    last_char=0;
    last_class=0;
    last_punct=0;
    
    gs=g_string_new("");
    if (last_dash) {
	g_string_append_c(gs,last_dash);
    }
    
    auto void add_punct(int nc)
    {
	if (!last_punct) return;
	if (last_punct & PUNCT_QUESTION) g_string_append_c(gs,'?');
	else if (last_punct & PUNCT_EXCLAM) g_string_append_c(gs,'!');
	else if (last_punct & PUNCT_DOT) g_string_append_c(gs,'.');
	else if (last_punct & PUNCT_COMMA) {
	    g_string_append_c(gs,nc?',':'.');
	}
	else if (!nc) {
	    g_string_append_c(gs,'.');
	}
	if (nc) g_string_append_c(gs,' ');
	last_punct=0;
    }
    
    auto void get_next_char(void)
    {
	last_char=this_char;
	last_class=this_class;
	this_char=next_char;
	this_class=next_class;
	next_char=get_char(&str);
	next_class=get_class(next_char);
    }
    
    
    for (;this_char;get_next_char()) {
	if (this_char == ' ') {
	    last_punct |= PUNCT_BLANK;
	    continue;
	}
	if (c_readable(this_class)) {
	    add_punct(this_char);
	    g_string_append_unichar(gs,this_char);
	    continue;
	}
	// spanish inverts after non-readable before alnum
	if ((this_char == 0xa1 || this_char == 0xbf) && 
	    !c_readable(last_class) && c_readable(next_class)) {
	    g_string_append_unichar(gs,this_char);
	    continue;
	}
	
	// dot, colon or comma between digits?
	if ((this_char == ',' || this_char =='.' || this_char == ':') &&
	    c_ascnum(last_class) && c_ascnum(next_class)) {
		g_string_append_unichar(gs,this_char);
		continue;
	}
	//dot between alnum? probably domain name...
	if (this_char == '.' &&
	    c_alnum(last_class) && c_alnum(next_class)) {
		last_punct |= PUNCT_BLANK;
		continue;
	}
	// dash or apostrophe between alnum?
	if ((this_char == '-' || this_char == '\'') &&
	    c_alnum(last_class) && c_alnum(next_class)) {
		g_string_append_unichar(gs,this_char);
		continue;
	}
	// dash after space and before digit?
	if (this_char == '-' && last_char == ' ' && c_ascnum(this_class)) {
		add_punct(this_char);
		g_string_append_unichar(gs,this_char);
		continue;
	}
	// dot after nonreadable before digit? Probably decimal...
	if (this_char == '.' && !c_readable(last_class) && c_ascnum(this_class)) {
	    add_punct(this_char);
	    g_string_append(gs,"0.");
	    continue;
	}
	// other punctuation?
	if (c_punct(this_class)) {
	    last_punct |= get_punct(this_char);
	    continue;
	}
	// ignore this character
	last_punct |= PUNCT_BLANK;
    }
    last_punct |= PUNCT_DOT;
    add_punct(0);
    return g_string_free(gs,FALSE);
}
	 
#ifdef TESTY
#include <stdio.h>
main(int argc,char *argv[])
{
    printf("%s\n",simplify_text(argv[1]));
}
#endif
