/*
 * subtitler.c - SubAloud subtitles reader
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <unistd.h>
#include "videosa.h"
#include "sonic.h"

void free_subtitle(struct subtitle *s)
{
    if (s->wave) g_free(s->wave);
    if (s->real_wave) g_free(s->real_wave);
    if (s->origin_text) g_free(s->origin_text);
    if (s->ortofn_text) g_free(s->ortofn_text);
    g_free(s);
}
    

void join_subtitles(void)
{
    if (!(module_feat & VSA_MFEAT_JOIN)) return;
    printf("Joining\n");
}

void preproc_subtitles(void)
{
    struct subtitle *s,**ss;
    char *c;
    if (module_feat & VSA_MFEAT_TRANS) {
	ss=&sublist;
	while (*ss) {
	    c=get_translation((*ss)->origin_text);
	    if (!c || !*c) {
		s=*ss;
		*ss=s->next;
		free_subtitle(s);
		if (c) g_free(c);
		continue;
	    }
	    (*ss)->ortofn_text=c;
	    ss=&(*ss)->next;
	}
    }
    join_subtitles();
}

int find_subtitle(void)
{
    int nsub;
    int curdec=current_decisec-sub_delay;
    if (curdec < 0) curdec=0;
    if (subarray_pos >=0 && curdec < subarray[subarray_pos]->decisecs) {
	subarray_pos=-1;
    }
    if (subarray_pos < 0) {
	for (nsub=0;nsub < subarray_len;nsub++) {
	    if (subarray[nsub]->decisecs >= curdec) break;
	    subarray_pos=nsub;
	}
    }
    else nsub=subarray_pos+1; // next subtitle
    if (nsub >= subarray_len) return -1;
    if (subarray[nsub]->decisecs > curdec) return -1;
    if (subarray[nsub]->decisecs < curdec-30) {
	int n;
	for (n=nsub+1;n<subarray_len;n++) {
	    if (subarray[n]->decisecs>curdec+10) break;
	    nsub=n;
	}
    }
    subarray_pos=nsub;
    return nsub;
}

static GThread *fetcher_Thread;
static int samplefreq;
GMutex *fetcherMutex;
void semaUp(void)
{
    g_mutex_unlock(fetcherMutex);
}
void semaDown(void)
{
    g_mutex_lock(fetcherMutex);
}


gpointer fetcher_ThreadFunc(gpointer data)
{
    int done=1;
    int curdec;
    struct subtitle *s;
    short *wave;
    int len,freq;
    for(;;) {
	if (!done) usleep(100000);
	else if (done < 0) sleep(1);
	done=0;
	if (mplayer_status == VSFL_FINISHED) return NULL;
	if (mplayer_status != VSFL_RUNNING && mplayer_status != VSFL_STOPPED) {
	    continue;
	}
	curdec=current_decisec-sub_delay;
	if (curdec < 0) curdec=0;
	semaDown();
	for (s=sublist;s;s=s->next) {
	    if (s->decisecs < curdec -50 || s->decisecs > curdec + 300) {
		if (!VSAU_FLAG(s,VSAU_RELAX)) continue;
		short *w=s->wave;
		VSAU_SET(s,VSAU_EMPTY);
		semaUp();
		printf("Freeing %s\n",s->origin_text);
		free(w);
		done=1;
		semaDown();
		continue;
	    }
	    if (s->decisecs < curdec) continue;
	    if (!VSAU_FLAG(s,VSAU_EMPTY)) continue;
	    VSAU_SET(s,VSAU_FETCHING);
	    semaUp();
	    printf("Fetching %s\n",s->origin_text);
	    wave=get_wave(s->ortofn_text,&len,&freq);
	    semaDown();
	    if (!wave) {
		VSAU_SET(s,VSAU_EMPTY);
		done=-1;
		break;
	    }
	    if (!samplefreq) samplefreq=freq;
	    s->wave_length=len;
	    s->wave=wave;
	    VSAU_SET(s,VSAU_RELAX);
	}
	semaUp();
    }
    return NULL;
}

void start_fetcher(void)
{
#if !GLIB_CHECK_VERSION(2,32,0)
    g_thread_init(NULL);
    fetcherMutex=g_mutex_new();
    fetcher_Thread=g_thread_create(fetcher_ThreadFunc,NULL,TRUE,NULL);
#else
    fetcherMutex=malloc(sizeof(*fetcherMutex));
    g_mutex_init(fetcherMutex);
    fetcher_Thread=g_thread_new("vms_fetcher",fetcher_ThreadFunc,NULL);
#endif
}

/* audio player loop */

void play_subtitles(void)
{
    int playing=-1;
    int do_sleep=0;
    short *wave;
    int wavelen;
    double wavedsec,waveplace;
    int que;
    double sonic_rate;
    sonicStream sonic=NULL;
    short sonic_buf[VSA_MAX_WAVE_FRAGMENT];
    int sonic_pos;
    
    void start_play(que)
    {
	VSAU_SET(subarray[que],VSAU_PLAYING);
	wave=subarray[que]->wave;
	wavelen=subarray[que]->wave_length;
	playing=que;
	sonic_rate = min_rate;
	printf("Start playing %d %s\n",subarray[que]->decisecs,subarray[que]->origin_text);
	if (que < subarray_len -1) {
	    wavedsec=(sonic_rate * (wavelen * 10 + samplefreq-1))/samplefreq;
	    waveplace=subarray[que+1]->decisecs - subarray[que]->decisecs;
	    if (waveplace < wavedsec) {
		if (waveplace < 2) sonic_rate=max_rate;
		else {
		    sonic_rate = wavedsec/waveplace;
		    if (sonic_rate > max_rate) sonic_rate=max_rate;
		}
		if (sonic_rate < min_rate) sonic_rate=min_rate;
	    }
	}
	if (sonic_rate > 1.01) {
	    int n=4 * VSA_MAX_WAVE_FRAGMENT;
	    if (n > wavelen) n=wavelen;
	    sonic=sonicCreateStream(samplefreq,1);
	    sonicSetSpeed(sonic,sonic_rate);
	    // initial wave
	    sonicWriteShortToStream(sonic,wave,n);
	    wave += n;
	    wavelen -= n;
	    sonic_pos=0;
	}
	else {
	    sonic=NULL;
	}
    }
    
    int play_fragment(void)
    {
	// returns 1 if not finished, 0 at end, -1 at error
	
	
	int nsamp,rc,nd,nw;
	if (!sonic) {
	    if (wavelen <= 0) return 0;
	    nsamp=wavelen;
	    if (nsamp > VSA_MAX_WAVE_FRAGMENT) nsamp = VSA_MAX_WAVE_FRAGMENT;
	    memcpy(sonic_buf,wave,2*nsamp);
	    volumize(sonic_buf,nsamp);
	    rc=audio_play_sfr(sonic_buf,nsamp,samplefreq);
	    wave += nsamp;
	    wavelen -= nsamp;
	    if (!rc) return -1; // error
	    return 1;
	}
	nsamp=sonicReadShortFromStream(sonic,sonic_buf,VSA_MAX_WAVE_FRAGMENT);
	if (nsamp < 0) return -1;
	sonic_pos = nsamp;
	if (sonic_pos < VSA_MAX_WAVE_FRAGMENT) {
	    if (wavelen <= 0) {
		sonicDestroyStream(sonic);
		sonic=NULL;
	    }
	    else {
		nw=4 * VSA_MAX_WAVE_FRAGMENT;
		if (nw > wavelen) nw=wavelen;
		sonicWriteShortToStream(sonic,wave,nw);
		wave += nw;
		wavelen -= nw;
		nd=VSA_MAX_WAVE_FRAGMENT-sonic_pos;
		nsamp=sonicReadShortFromStream(sonic,sonic_buf+sonic_pos,nd);
		if (nsamp < 0) return -1;
		sonic_pos += nsamp;
	    }
	}
	volumize(sonic_buf,sonic_pos);
	rc=audio_play_sfr(sonic_buf,sonic_pos,samplefreq);
	if (!rc) return -1; // error
	return 1;
    }
	
    
    for (;;) {
	if (synchro_sub) {
	    synchro_sub=0;
	    if (playing>=0) {
		audio_play_sfr(NULL,0,samplefreq);
		VSAU_SET(subarray[playing],VSAU_RELAX);
		if (sonic) {
		    sonicDestroyStream(sonic);
		    sonic=NULL;
		}
		playing=-1;
	    }
	    do_sleep=0;
	    subarray_pos=-1;
	}
	if (do_sleep) {
	    usleep(20000);
	    do_sleep=0;
	    continue;
	}
	if (mplayer_status == VSFL_FINISHED) break;
	if (mplayer_status != VSFL_RUNNING) {
	    subarray_pos = -1;
	    if (playing>=0) {
		audio_play_sfr(NULL,0,samplefreq);
		VSAU_SET(subarray[playing],VSAU_RELAX);
		if (sonic) {
		    sonicDestroyStream(sonic);
		    sonic=NULL;
		}
	    }
	    playing=-1;
	    do_sleep=1;
	    continue;
	}
	if (playing < 0) {
	    que=find_subtitle();
	    if (que < 0) {
		do_sleep=1;
		continue;
	    }
	    if (!VSAU_FLAG(subarray[que],VSAU_RELAX)) {
		do_sleep=1;
		continue;
	    }
	    start_play(que);
	}
	if (playing >= 0) {
	    int rc=play_fragment();
	    if (rc > 0) continue;
	    VSAU_SET(subarray[playing],VSAU_RELAX);
	    if (sonic) {
		sonicDestroyStream(sonic);
		sonic=NULL;
	    }
	    que=find_subtitle();
	    semaDown();
	    if (que < 0 || !VSAU_FLAG(subarray[que],VSAU_RELAX)) {
		semaUp();
		playing=-1;
		audio_play_sfr(NULL,0,samplefreq);
		do_sleep=1;
		continue;
	    }
	    start_play(que);
	    semaUp();
	}
    }
}

